import requests
from django.conf import settings
from django.contrib.auth.models import User
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from eo_gestion.eo_redmine.models import Project

REDMINE_HEADERS = {'X-Redmine-API-Key': settings.REDMINE_API_KEY}


def redmine_iterator(base_url, entity_name):
    result = requests.get(base_url + '?limit=100', headers=REDMINE_HEADERS).json()
    total = 0
    while total < result['total_count']:
        for item in result[entity_name]:
            yield item
            total += 1
        if total < result['total_count']:
            result = requests.get(base_url + '?limit=100&offset=%s' % total, headers=REDMINE_HEADERS).json()


redmine_user_cache = {}


def redmine_get_login(user_id):
    if not user_id in redmine_user_cache:
        user_data = requests.get(
            settings.REDMINE_BASE_URL + 'users/%s.json' % user_id, headers=REDMINE_HEADERS
        ).json()
        redmine_user_cache[user_id] = user_data['user']['login']
    return redmine_user_cache[user_id]


class Command(BaseCommand):
    @transaction.atomic
    def handle(self, *args, **options):
        alive_projects = []
        for project in redmine_iterator(settings.REDMINE_BASE_URL + 'projects.json', 'projects'):
            project_id = project['id']
            project_name = project['name']

            # Skip inactive or non-client projects
            if project['status'] != 1:
                continue
            if not ('parent' in project) or project['parent']['name'] != 'Projets Clients':
                continue
            alive_projects.append(project_id)

            # get the members
            cpf_ids = set()
            for membership in redmine_iterator(
                settings.REDMINE_BASE_URL + 'projects/%s/memberships.json' % project_id, 'memberships'
            ):
                if not 'user' in membership:
                    continue
                for role in membership['roles']:
                    if role['name'] == 'CPF':
                        cpf_ids.add(membership['user']['id'])
            cpfs = []
            for cpf_id in cpf_ids:
                try:
                    cpfs.append(User.objects.get(username=redmine_get_login(cpf_id)))
                except User.DoesNotExist:
                    # missing user in barba, ignoring, will be fixed later once he is created
                    pass

            # now match in our DB, maybe update
            try:
                project_object = Project.objects.get(id=project_id)
            except Project.DoesNotExist:
                project_object = Project()
                project_object.id = project_id
                project_object.name = project_name
                project_object.save()

            project_object.name = project_name
            project_object.cpfs.set(cpfs)
            project_object.save()

        # Purge dead projects
        Project.objects.exclude(id__in=alive_projects).delete()
